@@ -49,11 +49,13 @@
 #include <sys/stat.h>
 #include <assert.h>
 #include <errno.h>
-#include <QtCore/QDebug>
+#include <QLoggingCategory>
 #include <QtCore/QDir>
 #include <QtCore/QFile>
 #include <QtCore/QSocketNotifier>
 #include <QtCore/QTimer>
+#include <QtCore/QThread>
+#include <QtCore/QThreadStorage>
 #include <QtCore/QCoreApplication>
 
 #include <qplatformdefs.h> // QT_LSTAT, QT_STAT, QT_STATBUF
@@ -81,17 +83,24 @@
 
 #endif // HAVE_SYS_INOTIFY_H
 
+Q_DECLARE_LOGGING_CATEGORY(KDIRWATCH)
+#if QT_VERSION >= QT_VERSION_CHECK(5, 4, 0)
+// logging category for this framework, default: log stuff >= warning
+Q_LOGGING_CATEGORY(KDIRWATCH, "kf5.kcoreaddons.kdirwatch", QtWarningMsg)
+#else
+Q_LOGGING_CATEGORY(KDIRWATCH, "kf5.kcoreaddons.kdirwatch")
+#endif
+
 // set this to true for much more verbose debug output
 static const bool s_verboseDebug = false;
 
-// The KDirWatchPrivate instance is refcounted, and deleted by the last KDirWatch instance
-static KDirWatchPrivate *dwp_self = 0;
+static QThreadStorage<KDirWatchPrivate *> dwp_self;
 static KDirWatchPrivate *createPrivate()
 {
-    if (!dwp_self) {
-        dwp_self = new KDirWatchPrivate;
+    if (!dwp_self.hasLocalData()) {
+        dwp_self.setLocalData(new KDirWatchPrivate);
     }
-    return dwp_self;
+    return dwp_self.localData();
 }
 
 // Convert a string into a watch Method
@@ -113,7 +122,6 @@ static KDirWatch::Method methodFromString(const QByteArray &method)
     }
 }
 
-#ifndef NDEBUG
 static const char *methodToString(KDirWatch::Method method)
 {
     switch (method) {
@@ -125,11 +133,10 @@ static const char *methodToString(KDirWatch::Method method)
         return "Stat";
     case KDirWatch::QFSWatch:
         return "QFSWatch";
-    default:
-        return "ERROR!";
     }
+    // not reached
+    return 0;
 }
-#endif
 
 static const char s_envNfsPoll[] = "KDIRWATCH_NFSPOLLINTERVAL";
 static const char s_envPoll[] = "KDIRWATCH_POLLINTERVAL";
@@ -168,12 +175,11 @@ KDirWatchPrivate::KDirWatchPrivate()
     : timer(),
       freq(3600000), // 1 hour as upper bound
       statEntries(0),
-      m_ref(0),
       delayRemove(false),
       rescan_all(false),
       rescan_timer()
 {
-    timer.setObjectName(QLatin1String("KDirWatchPrivate::timer"));
+    timer.setObjectName(QStringLiteral("KDirWatchPrivate::timer"));
     connect(&timer, SIGNAL(timeout()), this, SLOT(slotRescan()));
 
     m_nfsPollInterval = qEnvironmentVariableIsSet(s_envNfsPoll) ? qgetenv(s_envNfsPoll).toInt() : 5000;
@@ -188,23 +194,14 @@ KDirWatchPrivate::KDirWatchPrivate()
     availableMethods << "Stat";
 
     // used for FAM and inotify
-    rescan_timer.setObjectName(QString::fromLatin1("KDirWatchPrivate::rescan_timer"));
+    rescan_timer.setObjectName(QStringLiteral("KDirWatchPrivate::rescan_timer"));
     rescan_timer.setSingleShot(true);
     connect(&rescan_timer, SIGNAL(timeout()), this, SLOT(slotRescan()));
 
 #if HAVE_FAM
-    // It's possible that FAM server can't be started
-    if (FAMOpen(&fc) == 0) {
-        availableMethods << "FAM";
-        use_fam = true;
-        sn = new QSocketNotifier(FAMCONNECTION_GETFD(&fc),
-                                 QSocketNotifier::Read, this);
-        connect(sn, SIGNAL(activated(int)),
-                this, SLOT(famEventReceived()));
-    } else {
-        qDebug() << "Can't use FAM (fam daemon not running?)";
-        use_fam = false;
-    }
+    availableMethods << "FAM";
+    use_fam = true;
+    sn = 0;
 #endif
 
 #if HAVE_SYS_INOTIFY_H
@@ -213,11 +210,11 @@ KDirWatchPrivate::KDirWatchPrivate()
     m_inotify_fd = inotify_init();
 
     if (m_inotify_fd <= 0) {
-        qDebug() << "Can't use Inotify, kernel doesn't support it";
+        qCDebug(KDIRWATCH) << "Can't use Inotify, kernel doesn't support it";
         supports_inotify = false;
     }
 
-    //qDebug() << "INotify available: " << supports_inotify;
+    //qCDebug(KDIRWATCH) << "INotify available: " << supports_inotify;
     if (supports_inotify) {
         availableMethods << "INotify";
         (void)fcntl(m_inotify_fd, F_SETFD, FD_CLOEXEC);
@@ -231,12 +228,13 @@ KDirWatchPrivate::KDirWatchPrivate()
     availableMethods << "QFileSystemWatcher";
     fsWatcher = 0;
 #endif
-#ifndef NDEBUG
-    qDebug() << "Available methods: " << availableMethods << "preferred=" << methodToString(m_preferredMethod);
-#endif
+
+    if (s_verboseDebug) {
+        qCDebug(KDIRWATCH) << "Available methods: " << availableMethods << "preferred=" << methodToString(m_preferredMethod);
+    }
 }
 
-// This is called on app exit (when K_GLOBAL_STATIC deletes KDirWatch::self)
+// This is called on app exit (deleted by QThreadStorage)
 KDirWatchPrivate::~KDirWatchPrivate()
 {
     timer.stop();
@@ -245,7 +243,7 @@ KDirWatchPrivate::~KDirWatchPrivate()
     removeEntries(0);
 
 #if HAVE_FAM
-    if (use_fam) {
+    if (use_fam && sn) {
         FAMClose(&fc);
     }
 #endif
@@ -261,7 +259,7 @@ KDirWatchPrivate::~KDirWatchPrivate()
 
 void KDirWatchPrivate::inotifyEventReceived()
 {
-    //qDebug();
+    //qCDebug(KDIRWATCH);
 #if HAVE_SYS_INOTIFY_H
     if (!supports_inotify) {
         return;
@@ -275,7 +273,7 @@ void KDirWatchPrivate::inotifyEventReceived()
 
     while (pending > 0) {
 
-        const int bytesToRead = qMin(pending, (int)sizeof(buf) - offsetStartRead);
+        const int bytesToRead = qMin<int>(pending, sizeof(buf) - offsetStartRead);
 
         int bytesAvailable = read(m_inotify_fd, &buf[offsetStartRead], bytesToRead);
         pending -= bytesAvailable;
@@ -283,8 +281,8 @@ void KDirWatchPrivate::inotifyEventReceived()
         offsetStartRead = 0;
 
         int offsetCurrent = 0;
-        while (bytesAvailable >= (int)sizeof(struct inotify_event)) {
-            const struct inotify_event *const event = (struct inotify_event *) &buf[offsetCurrent];
+        while (bytesAvailable >= int(sizeof(struct inotify_event))) {
+            const struct inotify_event *const event = reinterpret_cast<inotify_event *>(&buf[offsetCurrent]);
             const int eventSize = sizeof(struct inotify_event) + event->len;
             if (bytesAvailable < eventSize) {
                 break;
@@ -294,12 +292,18 @@ void KDirWatchPrivate::inotifyEventReceived()
             offsetCurrent += eventSize;
 
             QString path;
-            QByteArray cpath(event->name, event->len);
-            if (event->len) {
+            // strip trailing null chars, see inotify_event documentation
+            // these must not end up in the final QString version of path
+            int len = event->len;
+            while (len > 1 && !event->name[len - 1]) {
+                --len;
+            }
+            QByteArray cpath(event->name, len);
+            if (len) {
                 path = QFile::decodeName(cpath);
             }
 
-            if (path.length() && isNoisyFile(cpath.data())) {
+            if (!path.isEmpty() && isNoisyFile(cpath.data())) {
                 continue;
             }
 
@@ -321,12 +325,12 @@ void KDirWatchPrivate::inotifyEventReceived()
                     const QString tpath = e->path + QLatin1Char('/') + path;
 
                     //if (s_verboseDebug) {
-                    //  qDebug() << "got event" << "0x"+QString::number(event->mask, 16) << "for" << e->path;
+                    //  qCDebug(KDIRWATCH) << "got event" << "0x"+QString::number(event->mask, 16) << "for" << e->path;
                     //}
 
                     if (event->mask & IN_DELETE_SELF) {
                         if (s_verboseDebug) {
-                            qDebug() << "-->got deleteself signal for" << e->path;
+                            qCDebug(KDIRWATCH) << "-->got deleteself signal for" << e->path;
                         }
                         e->m_status = NonExistent;
                         e->wd = -1;
@@ -348,8 +352,8 @@ void KDirWatchPrivate::inotifyEventReceived()
                         Entry *sub_entry = e->findSubEntry(tpath);
 
                         if (s_verboseDebug) {
-                            qDebug() << "-->got CREATE signal for" << (tpath) << "sub_entry=" << sub_entry;
-                            qDebug() << *e;
+                            qCDebug(KDIRWATCH) << "-->got CREATE signal for" << (tpath) << "sub_entry=" << sub_entry;
+                            qCDebug(KDIRWATCH) << *e;
                         }
 
                         // The code below is very similar to the one in checkFAMEvent...
@@ -369,7 +373,7 @@ void KDirWatchPrivate::inotifyEventReceived()
                             }
                             if (!clients.isEmpty()) {
                                 emitEvent(e, Created, tpath);
-                                qDebug().nospace() << clients.count() << " instance(s) monitoring the new "
+                                qCDebug(KDIRWATCH).nospace() << clients.count() << " instance(s) monitoring the new "
                                                    << (isDir ? "dir " : "file ") << tpath;
                             }
                             e->m_pendingFileChanges.append(e->path);
@@ -380,7 +384,7 @@ void KDirWatchPrivate::inotifyEventReceived()
                     }
                     if (event->mask & (IN_DELETE | IN_MOVED_FROM)) {
                         if (s_verboseDebug) {
-                            qDebug() << "-->got DELETE signal for" << tpath;
+                            qCDebug(KDIRWATCH) << "-->got DELETE signal for" << tpath;
                         }
                         if ((e->isDir) && (!e->m_clients.empty())) {
                             Client *client = 0;
@@ -403,7 +407,7 @@ void KDirWatchPrivate::inotifyEventReceived()
                     if (event->mask & (IN_MODIFY | IN_ATTRIB)) {
                         if ((e->isDir) && (!e->m_clients.empty())) {
                             if (s_verboseDebug) {
-                                qDebug() << "-->got MODIFY signal for" << (tpath);
+                                qCDebug(KDIRWATCH) << "-->got MODIFY signal for" << (tpath);
                             }
                             // A file in this directory has been changed.  No
                             // addEntry/ removeEntry bookkeeping should be required.
@@ -532,7 +536,7 @@ QList<KDirWatchPrivate::Client *> KDirWatchPrivate::Entry::clientsForFileOrDir(c
         }
     } else {
         // Happens frequently, e.g. ERROR: couldn't stat "/home/dfaure/.viminfo.tmp"
-        //qDebug() << "ERROR: couldn't stat" << tpath;
+        //qCDebug(KDIRWATCH) << "ERROR: couldn't stat" << tpath;
         // In this case isDir is not set, but ret is empty anyway
         // so isDir won't be used.
     }
@@ -612,7 +616,7 @@ void KDirWatchPrivate::useFreq(Entry *e, int newFreq)
         if (timer.isActive()) {
             timer.start(freq);
         }
-        qDebug() << "Global Poll Freq is now" << freq << "msec";
+        qCDebug(KDIRWATCH) << "Global Poll Freq is now" << freq << "msec";
     }
 }
 
@@ -624,12 +628,27 @@ bool KDirWatchPrivate::useFAM(Entry *e)
         return false;
     }
 
+    if (!sn) {
+        if (FAMOpen(&fc) == 0) {
+            sn = new QSocketNotifier(FAMCONNECTION_GETFD(&fc),
+                                    QSocketNotifier::Read, this);
+            connect(sn, SIGNAL(activated(int)),
+                    this, SLOT(famEventReceived()));
+        } else {
+            use_fam = false;
+            return false;
+        }
+    }
+
     // handle FAM events to avoid deadlock
     // (FAM sends back all files in a directory when monitoring)
     famEventReceived();
 
     e->m_mode = FAMMode;
     e->dirty = false;
+    e->m_famReportedSeen = false;
+
+    bool startedFAMMonitor = false;
 
     if (e->isDir) {
         if (e->m_status == NonExistent) {
@@ -638,6 +657,7 @@ bool KDirWatchPrivate::useFAM(Entry *e)
         } else {
             int res = FAMMonitorDirectory(&fc, QFile::encodeName(e->path).data(),
                                           &(e->fr), e);
+            startedFAMMonitor = true;
             if (res < 0) {
                 e->m_mode = UnknownMode;
                 use_fam = false;
@@ -645,7 +665,7 @@ bool KDirWatchPrivate::useFAM(Entry *e)
                 sn = 0;
                 return false;
             }
-            qDebug().nospace() << " Setup FAM (Req " << FAMREQUEST_GETREQNUM(&(e->fr))
+            qCDebug(KDIRWATCH).nospace() << " Setup FAM (Req " << FAMREQUEST_GETREQNUM(&(e->fr))
                                << ") for " << e->path;
         }
     } else {
@@ -655,6 +675,7 @@ bool KDirWatchPrivate::useFAM(Entry *e)
         } else {
             int res = FAMMonitorFile(&fc, QFile::encodeName(e->path).data(),
                                      &(e->fr), e);
+            startedFAMMonitor = true;
             if (res < 0) {
                 e->m_mode = UnknownMode;
                 use_fam = false;
@@ -663,14 +684,21 @@ bool KDirWatchPrivate::useFAM(Entry *e)
                 return false;
             }
 
-            qDebug().nospace() << " Setup FAM (Req " << FAMREQUEST_GETREQNUM(&(e->fr))
+            qCDebug(KDIRWATCH).nospace() << " Setup FAM (Req " << FAMREQUEST_GETREQNUM(&(e->fr))
                                << ") for " << e->path;
         }
     }
 
     // handle FAM events to avoid deadlock
     // (FAM sends back all files in a directory when monitoring)
-    famEventReceived();
+    do {
+      famEventReceived();
+      if (startedFAMMonitor && !e->m_famReportedSeen) {
+        // 50 is ~half the time it takes to setup a watch.  If gamin's latency
+        // gets better, this can be reduced.
+        QThread::msleep(50);
+      }
+    } while (startedFAMMonitor &&!e->m_famReportedSeen);
 
     return true;
 }
@@ -680,7 +708,7 @@ bool KDirWatchPrivate::useFAM(Entry *e)
 // setup INotify notification, returns false if not possible
 bool KDirWatchPrivate::useINotify(Entry *e)
 {
-    //qDebug() << "trying to use inotify for monitoring";
+    //qCDebug(KDIRWATCH) << "trying to use inotify for monitoring";
 
     e->wd = -1;
     e->dirty = false;
@@ -702,12 +730,12 @@ bool KDirWatchPrivate::useINotify(Entry *e)
     if ((e->wd = inotify_add_watch(m_inotify_fd,
                                    QFile::encodeName(e->path).data(), mask)) >= 0) {
         if (s_verboseDebug) {
-            qDebug() << "inotify successfully used for monitoring" << e->path << "wd=" << e->wd;
+            qCDebug(KDIRWATCH) << "inotify successfully used for monitoring" << e->path << "wd=" << e->wd;
         }
         return true;
     }
 
-    qDebug() << "inotify failed for monitoring" << e->path << ":" << strerror(errno);
+    qCDebug(KDIRWATCH) << "inotify failed for monitoring" << e->path << ":" << strerror(errno);
     return false;
 }
 #endif
@@ -722,9 +750,9 @@ bool KDirWatchPrivate::useQFSWatch(Entry *e)
         return true;
     }
 
-    qDebug() << "fsWatcher->addPath" << e->path;
+    qCDebug(KDIRWATCH) << "fsWatcher->addPath" << e->path;
     if (!fsWatcher) {
-        fsWatcher = new KFileSystemWatcher();
+        fsWatcher = new QFileSystemWatcher();
         connect(fsWatcher, SIGNAL(directoryChanged(QString)), this, SLOT(fswEventReceived(QString)));
         connect(fsWatcher, SIGNAL(fileChanged(QString)),      this, SLOT(fswEventReceived(QString)));
     }
@@ -748,11 +776,11 @@ bool KDirWatchPrivate::useStat(Entry *e)
         if (statEntries == 1) {
             // if this was first STAT entry (=timer was stopped)
             timer.start(freq);      // then start the timer
-            qDebug() << " Started Polling Timer, freq " << freq;
+            qCDebug(KDIRWATCH) << " Started Polling Timer, freq " << freq;
         }
     }
 
-    qDebug() << " Setup Stat (freq " << e->freq << ") for " << e->path;
+    qCDebug(KDIRWATCH) << " Setup Stat (freq " << e->freq << ") for " << e->path;
 
     return true;
 }
@@ -769,7 +797,8 @@ void KDirWatchPrivate::addEntry(KDirWatch *instance, const QString &_path,
     if (path.isEmpty()
 #ifndef Q_OS_WIN
             || path == QLatin1String("/dev")
-            || (path.startsWith(QLatin1String("/dev/")) && !path.startsWith(QLatin1String("/dev/.")))
+            || (path.startsWith(QLatin1String("/dev/")) && !path.startsWith(QLatin1String("/dev/."))
+                && !path.startsWith(QLatin1String("/dev/shm")))
 #endif
        ) {
         return;    // Don't even go there.
@@ -784,7 +813,7 @@ void KDirWatchPrivate::addEntry(KDirWatch *instance, const QString &_path,
         if (sub_entry) {
             (*it).m_entries.append(sub_entry);
             if (s_verboseDebug) {
-                qDebug() << "Added already watched Entry" << path
+                qCDebug(KDIRWATCH) << "Added already watched Entry" << path
                          << "(for" << sub_entry->path << ")";
             }
 #if HAVE_SYS_INOTIFY_H
@@ -806,9 +835,9 @@ void KDirWatchPrivate::addEntry(KDirWatch *instance, const QString &_path,
         } else {
             (*it).addClient(instance, watchModes);
             if (s_verboseDebug) {
-                qDebug() << "Added already watched Entry" << path
+                qCDebug(KDIRWATCH) << "Added already watched Entry" << path
                          << "(now" << (*it).clientCount() << "clients)"
-                         << QString::fromLatin1("[%1]").arg(instance->objectName());
+                         << QStringLiteral("[%1]").arg(instance->objectName());
             }
         }
         return;
@@ -873,10 +902,12 @@ void KDirWatchPrivate::addEntry(KDirWatch *instance, const QString &_path,
         e->addClient(instance, watchModes);
     }
 
-    qDebug().nospace() << "Added " << (e->isDir ? "Dir " : "File ") << path
+    if (s_verboseDebug) {
+        qCDebug(KDIRWATCH).nospace() << "Added " << (e->isDir ? "Dir " : "File ") << path
                        << (e->m_status == NonExistent ? " NotExisting" : "")
                        << " for " << (sub_entry ? sub_entry->path : QString())
                        << " [" << (instance ? instance->objectName() : QString()) << "]";
+    }
 
     // now setup the notification method
     e->m_mode = UnknownMode;
@@ -900,7 +931,7 @@ void KDirWatchPrivate::addEntry(KDirWatch *instance, const QString &_path,
 
 #if HAVE_SYS_INOTIFY_H
         if (e->m_mode == INotifyMode || (e->m_mode == UnknownMode && m_preferredMethod == KDirWatch::INotify)) {
-            //qDebug() << "Ignoring WatchFiles directive - this is implicit with inotify";
+            //qCDebug(KDIRWATCH) << "Ignoring WatchFiles directive - this is implicit with inotify";
             // Placing a watch on individual files is redundant with inotify
             // (inotify gives us WatchFiles functionality "for free") and indeed
             // actively harmful, so prevent it.  WatchSubDirs is necessary, though.
@@ -954,7 +985,6 @@ void KDirWatchPrivate::addWatch(Entry *e)
     case KDirWatch::QFSWatch: entryAdded = useQFSWatch(e); break;
 #endif
     case KDirWatch::Stat: entryAdded = useStat(e); break;
-    default: break;
     }
 
     // Failing that try in order INotify, FAM, QFSWatch, Stat
@@ -983,7 +1013,7 @@ void KDirWatchPrivate::removeWatch(Entry *e)
 #if HAVE_FAM
     if (e->m_mode == FAMMode) {
         FAMCancelMonitor(&fc, &(e->fr));
-        qDebug().nospace()  << "Cancelled FAM (Req " << FAMREQUEST_GETREQNUM(&(e->fr))
+        qCDebug(KDIRWATCH).nospace()  << "Cancelled FAM (Req " << FAMREQUEST_GETREQNUM(&(e->fr))
                             << ") for " << e->path;
     }
 #endif
@@ -991,7 +1021,7 @@ void KDirWatchPrivate::removeWatch(Entry *e)
     if (e->m_mode == INotifyMode) {
         (void) inotify_rm_watch(m_inotify_fd, e->wd);
         if (s_verboseDebug) {
-            qDebug().nospace() << "Cancelled INotify (fd " << m_inotify_fd << ", "
+            qCDebug(KDIRWATCH).nospace() << "Cancelled INotify (fd " << m_inotify_fd << ", "
                                << e->wd << ") for " << e->path;
         }
     }
@@ -999,7 +1029,7 @@ void KDirWatchPrivate::removeWatch(Entry *e)
 #if HAVE_QFILESYSTEMWATCHER
     if (e->m_mode == QFSWatchMode && fsWatcher) {
         if (s_verboseDebug) {
-            qDebug() << "fsWatcher->removePath" << e->path;
+            qCDebug(KDIRWATCH) << "fsWatcher->removePath" << e->path;
         }
         fsWatcher->removePath(e->path);
     }
@@ -1011,7 +1041,7 @@ bool KDirWatchPrivate::removeEntry(KDirWatch *instance,
                                    Entry *sub_entry)
 {
     if (s_verboseDebug) {
-        qDebug() << "path=" << _path << "sub_entry:" << sub_entry;
+        qCDebug(KDIRWATCH) << "path=" << _path << "sub_entry:" << sub_entry;
     }
     Entry *e = entry(_path);
     if (!e) {
@@ -1034,7 +1064,7 @@ void KDirWatchPrivate::removeEntry(KDirWatch *instance,
         e->removeClient(instance);
     }
 
-    if (e->m_clients.count() || e->m_entries.count()) {
+    if (!e->m_clients.empty() || !e->m_entries.empty()) {
         return;
     }
 
@@ -1059,12 +1089,12 @@ void KDirWatchPrivate::removeEntry(KDirWatch *instance,
         statEntries--;
         if (statEntries == 0) {
             timer.stop(); // stop timer if lists are empty
-            qDebug() << " Stopped Polling Timer";
+            qCDebug(KDIRWATCH) << " Stopped Polling Timer";
         }
     }
 
     if (s_verboseDebug) {
-        qDebug().nospace() << "Removed " << (e->isDir ? "Dir " : "File ") << e->path
+        qCDebug(KDIRWATCH).nospace() << "Removed " << (e->isDir ? "Dir " : "File ") << e->path
                            << " for " << (sub_entry ? sub_entry->path : QString())
                            << " [" << (instance ? instance->objectName() : QString()) << "]";
     }
@@ -1108,7 +1138,7 @@ void KDirWatchPrivate::removeEntries(KDirWatch *instance)
         if (timer.isActive()) {
             timer.start(freq);
         }
-        qDebug() << "Poll Freq now" << freq << "msec";
+        qCDebug(KDIRWATCH) << "Poll Freq now" << freq << "msec";
     }
 }
 
@@ -1124,7 +1154,7 @@ bool KDirWatchPrivate::stopEntryScan(KDirWatch *instance, Entry *e)
         }
     }
 
-    qDebug() << (instance ? instance->objectName() : QString::fromLatin1("all"))
+    qCDebug(KDIRWATCH) << (instance ? instance->objectName() : QStringLiteral("all"))
              << "stopped scanning" << e->path << "(now"
              << stillWatching << "watchers)";
 
@@ -1157,7 +1187,7 @@ bool KDirWatchPrivate::restartEntryScan(KDirWatch *instance, Entry *e,
         return false;
     }
 
-    qDebug() << (instance ? instance->objectName() : QString::fromLatin1("all"))
+    qCDebug(KDIRWATCH) << (instance ? instance->objectName() : QStringLiteral("all"))
              << "restarted scanning" << e->path
              << "(now" << wasWatching + newWatching << "watchers)";
 
@@ -1174,7 +1204,7 @@ bool KDirWatchPrivate::restartEntryScan(KDirWatch *instance, Entry *e,
                 e->m_ctime = qMax(stat_buf.st_ctime, stat_buf.st_mtime);
                 e->m_status = Normal;
                 if (s_verboseDebug) {
-                    qDebug() << "Setting status to Normal for" << e << e->path;
+                    qCDebug(KDIRWATCH) << "Setting status to Normal for" << e << e->path;
                 }
                 e->m_nlink = stat_buf.st_nlink;
                 e->m_ino = stat_buf.st_ino;
@@ -1186,7 +1216,7 @@ bool KDirWatchPrivate::restartEntryScan(KDirWatch *instance, Entry *e,
                 e->m_status = NonExistent;
                 e->m_nlink = 0;
                 if (s_verboseDebug) {
-                    qDebug() << "Setting status to NonExistent for" << e << e->path;
+                    qCDebug(KDIRWATCH) << "Setting status to NonExistent for" << e << e->path;
                 }
             }
         }
@@ -1277,7 +1307,7 @@ int KDirWatchPrivate::scanEntry(Entry *e)
             e->m_status = Normal;
             e->m_ino = stat_buf.st_ino;
             if (s_verboseDebug) {
-                qDebug() << "Setting status to Normal for just created" << e << e->path;
+                qCDebug(KDIRWATCH) << "Setting status to Normal for just created" << e << e->path;
             }
             // We need to make sure the entry isn't listed in its parent's subentries... (#222974, testMoveTo)
             removeEntry(0, e->parentDirectory(), e);
@@ -1290,7 +1320,7 @@ int KDirWatchPrivate::scanEntry(Entry *e)
             struct tm *tmp = localtime(&e->m_ctime);
             char outstr[200];
             strftime(outstr, sizeof(outstr), "%H:%M:%S", tmp);
-            qDebug() << e->path << "e->m_ctime=" << e->m_ctime << outstr
+            qCDebug(KDIRWATCH) << e->path << "e->m_ctime=" << e->m_ctime << outstr
                      << "stat_buf.st_ctime=" << stat_buf.st_ctime
                      << "stat_buf.st_mtime=" << stat_buf.st_mtime
                      << "e->m_nlink=" << e->m_nlink
@@ -1317,9 +1347,11 @@ int KDirWatchPrivate::scanEntry(Entry *e)
                 // The file got deleted and recreated. We need to watch it again.
                 removeWatch(e);
                 addWatch(e);
+                e->m_ino = stat_buf.st_ino;
+                return (Deleted|Created);
+            } else {
+              return Changed;
             }
-            e->m_ino = stat_buf.st_ino;
-            return Changed;
         }
 
         return NoChange;
@@ -1360,7 +1392,7 @@ void KDirWatchPrivate::emitEvent(const Entry *e, int event, const QString &fileN
     }
 
     if (s_verboseDebug) {
-        qDebug() << event << path << e->m_clients.count() << "clients";
+        qCDebug(KDIRWATCH) << event << path << e->m_clients.count() << "clients";
     }
 
     Q_FOREACH (Client *c, e->m_clients) {
@@ -1392,8 +1424,6 @@ void KDirWatchPrivate::emitEvent(const Entry *e, int event, const QString &fileN
 
         if (event & Deleted) {
             QMetaObject::invokeMethod(c->instance, "setDeleted", Qt::QueuedConnection, Q_ARG(QString, path));
-            // emit only Deleted event...
-            continue;
         }
 
         if (event & Created) {
@@ -1426,7 +1456,7 @@ void KDirWatchPrivate::slotRemoveDelayed()
 void KDirWatchPrivate::slotRescan()
 {
     if (s_verboseDebug) {
-        qDebug();
+        qCDebug(KDIRWATCH);
     }
 
     EntryMap::Iterator it;
@@ -1475,7 +1505,7 @@ void KDirWatchPrivate::slotRescan()
 
         const int ev = scanEntry(entry);
         if (s_verboseDebug) {
-            qDebug() << "scanEntry for" << entry->path << "says" << ev;
+            qCDebug(KDIRWATCH) << "scanEntry for" << entry->path << "says" << ev;
         }
 
         switch (entry->m_mode) {
@@ -1483,12 +1513,12 @@ void KDirWatchPrivate::slotRescan()
         case INotifyMode:
             if (ev == Deleted) {
                 if (s_verboseDebug) {
-                    qDebug() << "scanEntry says" << entry->path << "was deleted";
+                    qCDebug(KDIRWATCH) << "scanEntry says" << entry->path << "was deleted";
                 }
                 addEntry(0, entry->parentDirectory(), entry, true);
             } else if (ev == Created) {
                 if (s_verboseDebug) {
-                    qDebug() << "scanEntry says" << entry->path << "was created. wd=" << entry->wd;
+                    qCDebug(KDIRWATCH) << "scanEntry says" << entry->path << "was created. wd=" << entry->wd;
                 }
                 if (entry->wd < 0) {
                     cList.append(entry);
@@ -1518,7 +1548,7 @@ void KDirWatchPrivate::slotRescan()
             pendingFileChanges.removeDuplicates();
             Q_FOREACH (const QString &changedFilename, pendingFileChanges) {
                 if (s_verboseDebug) {
-                    qDebug() << "processing pending file change for" << changedFilename;
+                    qCDebug(KDIRWATCH) << "processing pending file change for" << changedFilename;
                 }
                 emitEvent(entry, Changed, changedFilename);
             }
@@ -1572,7 +1602,7 @@ void KDirWatchPrivate::famEventReceived()
 
     delayRemove = true;
 
-    //qDebug() << "Fam event received";
+    //qCDebug(KDIRWATCH) << "Fam event received";
 
     while (use_fam && FAMPending(&fc)) {
         if (FAMNextEvent(&fc, &fe) == -1) {
@@ -1597,18 +1627,7 @@ void KDirWatchPrivate::famEventReceived()
 
 void KDirWatchPrivate::checkFAMEvent(FAMEvent *fe)
 {
-    //qDebug();
-
-    // Don't be too verbose ;-)
-    if ((fe->code == FAMExists) ||
-            (fe->code == FAMEndExist) ||
-            (fe->code == FAMAcknowledge)) {
-        return;
-    }
-
-    if (isNoisyFile(fe->filename)) {
-        return;
-    }
+    //qCDebug(KDIRWATCH);
 
     Entry *e = 0;
     EntryMap::Iterator it = m_mapEntries.begin();
@@ -1619,10 +1638,24 @@ void KDirWatchPrivate::checkFAMEvent(FAMEvent *fe)
             break;
         }
 
+    // Don't be too verbose ;-)
+    if ((fe->code == FAMExists) ||
+            (fe->code == FAMEndExist) ||
+            (fe->code == FAMAcknowledge)) {
+      if (e) {
+        e->m_famReportedSeen = true;
+      }
+      return;
+    }
+
+    if (isNoisyFile(fe->filename)) {
+        return;
+    }
+
     // Entry *e = static_cast<Entry*>(fe->userdata);
 
     if (s_verboseDebug) { // don't enable this except when debugging, see #88538
-        qDebug()  << "Processing FAM event ("
+        qCDebug(KDIRWATCH)  << "Processing FAM event ("
                   << ((fe->code == FAMChanged) ? "FAMChanged" :
                       (fe->code == FAMDeleted) ? "FAMDeleted" :
                       (fe->code == FAMStartExecuting) ? "FAMStartExecuting" :
@@ -1638,12 +1671,12 @@ void KDirWatchPrivate::checkFAMEvent(FAMEvent *fe)
 
     if (!e) {
         // this happens e.g. for FAMAcknowledge after deleting a dir...
-        //    qDebug() << "No entry for FAM event ?!";
+        //    qCDebug(KDIRWATCH) << "No entry for FAM event ?!";
         return;
     }
 
     if (e->m_status == NonExistent) {
-        qDebug() << "FAM event for nonExistent entry " << e->path;
+        qCDebug(KDIRWATCH) << "FAM event for nonExistent entry " << e->path;
         return;
     }
 
@@ -1659,7 +1692,7 @@ void KDirWatchPrivate::checkFAMEvent(FAMEvent *fe)
         // fe->filename is an absolute path when a watched file-or-dir is deleted
         if (!QDir::isRelativePath(QFile::decodeName(fe->filename))) {
             FAMCancelMonitor(&fc, &(e->fr)); // needed ?
-            qDebug()  << "Cancelled FAMReq"
+            qCDebug(KDIRWATCH)  << "Cancelled FAMReq"
                       << FAMREQUEST_GETREQNUM(&(e->fr))
                       << "for" << e->path;
             e->m_status = NonExistent;
@@ -1676,7 +1709,7 @@ void KDirWatchPrivate::checkFAMEvent(FAMEvent *fe)
             // A file in this directory has been removed, and wasn't explicitly watched.
             // We could still inform clients, like inotify does? But stat can't.
             // For now we just marked e dirty and slotRescan will emit the dir as dirty.
-            //qDebug() << "Got FAMDeleted for" << QFile::decodeName(fe->filename) << "in" << e->path << ". Absolute path -> NOOP!";
+            //qCDebug(KDIRWATCH) << "Got FAMDeleted for" << QFile::decodeName(fe->filename) << "in" << e->path << ". Absolute path -> NOOP!";
         }
         break;
 
@@ -1687,8 +1720,9 @@ void KDirWatchPrivate::checkFAMEvent(FAMEvent *fe)
         // This code is very similar to the one in inotifyEventReceived...
         Entry *sub_entry = e->findSubEntry(tpath);
         if (sub_entry /*&& sub_entry->isDir*/) {
-            // We were waiting for this new file/dir to be created
-            emitEvent(sub_entry, Created);
+            // We were waiting for this new file/dir to be created.  We don't actually
+            // emit an event here, as the rescan_timer will re-detect the creation and
+            // do the signal emission there.
             sub_entry->dirty = true;
             rescan_timer.start(0); // process this asap, to start watching that dir
         } else if (e->isDir && !e->m_clients.empty()) {
@@ -1702,7 +1736,7 @@ void KDirWatchPrivate::checkFAMEvent(FAMEvent *fe)
             if (!clients.isEmpty()) {
                 emitEvent(e, Created, tpath);
 
-                qDebug().nospace() << clients.count() << " instance(s) monitoring the new "
+                qCDebug(KDIRWATCH).nospace() << clients.count() << " instance(s) monitoring the new "
                                    << (isDir ? "dir " : "file ") << tpath;
             }
         }
@@ -1723,14 +1757,14 @@ void KDirWatchPrivate::statistics()
 {
     EntryMap::Iterator it;
 
-    qDebug() << "Entries watched:";
+    qCDebug(KDIRWATCH) << "Entries watched:";
     if (m_mapEntries.count() == 0) {
-        qDebug() << "  None.";
+        qCDebug(KDIRWATCH) << "  None.";
     } else {
         it = m_mapEntries.begin();
         for (; it != m_mapEntries.end(); ++it) {
             Entry *e = &(*it);
-            qDebug() << "  " << *e;
+            qCDebug(KDIRWATCH) << "  " << *e;
 
             Q_FOREACH (Client *c, e->m_clients) {
                 QByteArray pending;
@@ -1749,13 +1783,13 @@ void KDirWatchPrivate::statistics()
                     }
                     pending = ", stopped" + pending;
                 }
-                qDebug()  << "    by " << c->instance->objectName()
+                qCDebug(KDIRWATCH)  << "    by " << c->instance->objectName()
                           << " (" << c->count << " times)" << pending;
             }
             if (e->m_entries.count() > 0) {
-                qDebug() << "    dependent entries:";
+                qCDebug(KDIRWATCH) << "    dependent entries:";
                 Q_FOREACH (Entry *d, e->m_entries) {
-                    qDebug() << "      " << d << d->path << (d->m_status == NonExistent ? "NonExistent" : "EXISTS!!! ERROR!");
+                    qCDebug(KDIRWATCH) << "      " << d << d->path << (d->m_status == NonExistent ? "NonExistent" : "EXISTS!!! ERROR!");
                     if (s_verboseDebug) {
                         Q_ASSERT(d->m_status == NonExistent); // it doesn't belong here otherwise
                     }
@@ -1770,7 +1804,7 @@ void KDirWatchPrivate::statistics()
 void KDirWatchPrivate::fswEventReceived(const QString &path)
 {
     if (s_verboseDebug) {
-        qDebug() << path;
+        qCDebug(KDIRWATCH) << path;
     }
     EntryMap::Iterator it = m_mapEntries.find(path);
     if (it != m_mapEntries.end()) {
@@ -1778,7 +1812,7 @@ void KDirWatchPrivate::fswEventReceived(const QString &path)
         e->dirty = true;
         const int ev = scanEntry(e);
         if (s_verboseDebug) {
-            qDebug() << "scanEntry for" << e->path << "says" << ev;
+            qCDebug(KDIRWATCH) << "scanEntry for" << e->path << "says" << ev;
         }
         if (ev != NoChange) {
             emitEvent(e, ev);
@@ -1797,6 +1831,14 @@ void KDirWatchPrivate::fswEventReceived(const QString &path)
             Q_FOREACH (Entry *sub_entry, e->m_entries) {
                 fswEventReceived(sub_entry->path); // recurse, to call scanEntry and see if something changed
             }
+        } else {
+            /* Even though QFileSystemWatcher only reported the file as modified, it is possible that the file
+             * was in fact just deleted and then immediately recreated.  If the file was deleted, QFileSystemWatcher
+             * will delete the watch, and will ignore the file, even after it is recreated.  Since it is impossible
+             * to reliably detect this case, always re-request the watch on a dirty signal, to avoid losing the
+             * underlying OS monitor.
+             */
+            fsWatcher->addPath(e->path);
         }
     }
 }
@@ -1822,41 +1864,35 @@ KDirWatch *KDirWatch::self()
 // <dfaure> yes, see kio/src/core/kcoredirlister_p.h:328
 bool KDirWatch::exists()
 {
-    return s_pKDirWatchSelf.exists();
+    return s_pKDirWatchSelf.exists() && dwp_self.hasLocalData();
 }
 
-static void cleanupQFSWatcher()
+static void postRoutine_KDirWatch()
 {
-    s_pKDirWatchSelf()->deleteQFSWatcher();
+    if (s_pKDirWatchSelf.exists()) {
+        s_pKDirWatchSelf()->deleteQFSWatcher();
+    }
 }
 
 KDirWatch::KDirWatch(QObject *parent)
     : QObject(parent), d(createPrivate())
 {
-    static int nameCounter = 0;
-
-    nameCounter++;
-    setObjectName(QString::fromLatin1("KDirWatch-%1").arg(nameCounter));
-
-    d->ref();
+    static QBasicAtomicInt nameCounter = Q_BASIC_ATOMIC_INITIALIZER(1);
+    const int counter = nameCounter.fetchAndAddRelaxed(1); // returns the old value
+    setObjectName(QStringLiteral("KDirWatch-%1").arg(counter));
 
     d->_isStopped = false;
 
-    static bool cleanupRegistered = false;
-    if (!cleanupRegistered) {
-        cleanupRegistered = true;
+    if (counter == 1) { // very first KDirWatch instance
         // Must delete QFileSystemWatcher before qApp is gone - bug 261541
-        qAddPostRoutine(cleanupQFSWatcher);
+        qAddPostRoutine(postRoutine_KDirWatch);
     }
 }
 
 KDirWatch::~KDirWatch()
 {
-    d->removeEntries(this);
-    if (d->deref()) {
-        // delete it if it's the last one
-        delete d;
-        dwp_self = 0;
+    if (dwp_self.hasLocalData()) { // skip this after app destruction
+        d->removeEntries(this);
     }
 }
 
@@ -1970,32 +2006,33 @@ void KDirWatch::deleteQFSWatcher()
 {
     delete d->fsWatcher;
     d->fsWatcher = 0;
+    d = 0;
 }
 
 void KDirWatch::statistics()
 {
-    if (!dwp_self) {
-        qDebug() << "KDirWatch not used";
+    if (!dwp_self.hasLocalData()) {
+        qCDebug(KDIRWATCH) << "KDirWatch not used";
         return;
     }
-    dwp_self->statistics();
+    dwp_self.localData()->statistics();
 }
 
 void KDirWatch::setCreated(const QString &_file)
 {
-    qDebug() << objectName() << "emitting created" << _file;
+    qCDebug(KDIRWATCH) << objectName() << "emitting created" << _file;
     emit created(_file);
 }
 
 void KDirWatch::setDirty(const QString &_file)
 {
-    //qDebug() << objectName() << "emitting dirty" << _file;
+    //qCDebug(KDIRWATCH) << objectName() << "emitting dirty" << _file;
     emit dirty(_file);
 }
 
 void KDirWatch::setDeleted(const QString &_file)
 {
-    qDebug() << objectName() << "emitting deleted" << _file;
+    qCDebug(KDIRWATCH) << objectName() << "emitting deleted" << _file;
     emit deleted(_file);
 }
 
@@ -2007,17 +2044,18 @@ KDirWatch::Method KDirWatch::internalMethod() const
     case KDirWatch::FAM: if (d->use_fam) {
             return KDirWatch::FAM;
         }
+        break;
 #endif
 #if HAVE_SYS_INOTIFY_H
     case KDirWatch::INotify: if (d->supports_inotify) {
             return KDirWatch::INotify;
         }
+        break;
 #endif
 #if HAVE_QFILESYSTEMWATCHER
     case KDirWatch::QFSWatch: return KDirWatch::QFSWatch;
 #endif
     case KDirWatch::Stat: return KDirWatch::Stat;
-    default: break;
     }
 
 #if HAVE_SYS_INOTIFY_H
@@ -2032,8 +2070,9 @@ KDirWatch::Method KDirWatch::internalMethod() const
 #endif
 #if HAVE_QFILESYSTEMWATCHER
     return KDirWatch::QFSWatch;
-#endif
+#else
     return KDirWatch::Stat;
+#endif
 }
 
 #include "moc_kdirwatch.cpp"
